"""
    tests
    ~~~~~
"""

__all__ = ['fixture', 'TestCase']

import os.path
import unittest
import json

here = os.path.dirname(os.path.abspath(__file__))


def fixture(filename):
    fullpath = os.path.join(here, 'fixtures', filename)
    with open(fullpath, 'r') as file:
        return json.load(file)


class TestCase(unittest.TestCase):
    pass



"""
    json.schema.bases
    ~~~~~~~~~~~~~~~~~

"""

from __future__ import absolute_import, print_function, unicode_literals

__all__ = ['BaseValidator']

from abc import abstractmethod


class Validator(object):
    @abstractmethod
    def has_default(self):
        pass

    @abstractmethod
    def validate(self, obj):
        pass


class ReferenceValidator(Validator):
    """
    Loads lately validator.
    """

    def __init__(self, uri, factory, spec=None):
        self.uri = uri
        self.factory = factory
        self.spec = spec

    @property
    def validator(self):
        if not hasattr(self, '_validator'):
            self._validator = self.factory.load(self.uri, self.spec)

        return self._validator

    def has_default(self):
        return self.validator.has_default()

    @property
    def default(self):
        return self.validator.default

    def validate(self, obj):
        return self.validator.validate(obj)



"""
    tests.tests_schemas
    ~~~~~~~~~~~~~~~~~~~

    examples are taken from:

    * http://spacetelescope.github.io/understanding-json-schema
    * http://json-schema.org/examples.html
"""

from json.schema import load
from json.schema.exceptions import ValidationError
from . import TestCase, fixture


class TestSchema(TestCase):

    def test_first(self):
        data1 = fixture('first.data1.json')
        data2 = fixture('first.data2.json')
        schema = fixture('first.schema.json')

        validator = load(schema)
        with self.assertRaises(ValidationError):
            validator.validate(data1)
        validator.validate(data2)

    def test_second(self):
        schema = fixture('second.schema.json')
        data = fixture('second.data1.json')
        validator = load(schema)
        validator.validate(data)

    def test_three(self):
        schema = fixture('three.schema.json')
        data1 = fixture('three.data1.json')
        data2 = fixture('three.data2.json')
        validator = load(schema)

        assert validator.validate(data2) == {
            'shipping_address': {
                'street_address': '1600 Pennsylvania Avenue NW',
                'city': 'Washington',
                'state': 'DC',
                'type': 'business'
            }
        }

        with self.assertRaises(ValidationError):
            res = validator.validate(data1)
            print('-', res)

    def test_four(self):
        data = fixture('four.data.json')
        base_schema = fixture('four.base.schema.json')
        entry_schema = fixture('four.entry.schema.json')
        validator = load(base_schema, loader={
            'http://some.site.somewhere/entry-schema#': entry_schema
        })
        validator.validate(data)

    def test_five(self):
        validator = load(fixture('five.schema.json'))
        validator.validate({
            'creditcard': {
                'provider': 'visa',
                'securitycode': 123
            }
        })
        validator.validate({
            'creditcard': {
                'provider': 'mastercard'
            }
        })

        with self.assertRaises(ValidationError):
            validator.validate({
                'creditcard': {
                    'provider': 'visa'
                }
            })

        with self.assertRaises(ValidationError):
            validator.validate({
                'creditcard': {
                    'provider': 'mastercard',
                    'securitycode': 123
                }
            })



from __future__ import absolute_import, print_function, unicode_literals
from json import schema
from json.schema.exceptions import ValidationError
import os

here = os.path.dirname(os.path.abspath(__file__))

dataset = [
    ('basic.json', {
        'firstName': 'John',
        'lastName': 'Noone',
        'age': 33,
    }),
    ('simple.json', [{
        'id': 1,
        'name': 'A green door',
        'price': 12.50,
        'tags': ['home', 'green']
    }]),
]

for filename, data in dataset:
    with open(os.path.join(here, filename)) as file:
        print('validate {}'.format(filename))

        validator = schema.load(file)
        try:
            validator.validate(data)
            print('all fine')
        except ValidationError as error:
            print(error)



"""
    tests.tests_draft04
    ~~~~~~~~~~~~~~~~~~~

"""

from json.schema.draft04 import Draft04Validator
from json.schema.exceptions import ValidationError
from . import TestCase


class TestNumber(TestCase):

    def test_any_number(self):
        validator = Draft04Validator(type='number')
        validator.validate(1)
        validator.validate(1.1)
        validator.validate(-12e3)

        with self.assertRaises(ValidationError):
            validator.validate('foo')

        with self.assertRaises(ValidationError):
            validator.validate(True)

    def test_minimum(self):
        validator = Draft04Validator(type='number', minimum=2)
        validator.validate(2)
        validator.validate(2.0)

        with self.assertRaises(ValidationError):
            validator.validate(1.9)

        validator = Draft04Validator(type='number', minimum=2, exclusiveMinimum=True)
        with self.assertRaises(ValidationError):
            validator.validate(2)
        with self.assertRaises(ValidationError):
            validator.validate(2.0)
        with self.assertRaises(ValidationError):
            validator.validate(1.9)
        validator.validate(2.01)

    def test_maximum(self):
        validator = Draft04Validator(type='number', maximum=2)
        validator.validate(2)
        validator.validate(2.0)

        with self.assertRaises(ValidationError):
            validator.validate(2.1)

        validator = Draft04Validator(type='number', maximum=2, exclusiveMaximum=True)
        with self.assertRaises(ValidationError):
            validator.validate(2)
        with self.assertRaises(ValidationError):
            validator.validate(2.0)
        with self.assertRaises(ValidationError):
            validator.validate(2.1)
        validator.validate(1.9991)

    def test_multiple(self):
        validator = Draft04Validator(type='number', multipleOf=2)
        validator.validate(2)
        validator.validate(2.0)

        with self.assertRaises(ValidationError):
            validator.validate(3)

        validator = Draft04Validator(type='number', multipleOf=2.5)
        validator.validate(2.5)
        validator.validate(50)
        validator.validate(-50)

        with self.assertRaises(ValidationError):
            validator.validate(-51)


class TestInteger(TestCase):

    def test_any_integer(self):
        validator = Draft04Validator(type='integer')
        validator.validate(1)
        with self.assertRaises(ValidationError):
            validator.validate(1.1)

        with self.assertRaises(ValidationError):
            validator.validate(-12e3)

        with self.assertRaises(ValidationError):
            validator.validate('foo')

        with self.assertRaises(ValidationError):
            validator.validate(True)

    def test_minimum(self):
        validator = Draft04Validator(type='integer', minimum=2)
        validator.validate(2)
        with self.assertRaises(ValidationError):
            validator.validate(2.0)

        with self.assertRaises(ValidationError):
            validator.validate(1.9)

        validator = Draft04Validator(type='integer', minimum=2, exclusiveMinimum=True)
        with self.assertRaises(ValidationError):
            validator.validate(2)
        with self.assertRaises(ValidationError):
            validator.validate(2.0)
        with self.assertRaises(ValidationError):
            validator.validate(1.9)
        with self.assertRaises(ValidationError):
            validator.validate(2.01)

    def test_maximum(self):
        validator = Draft04Validator(type='integer', maximum=2)
        validator.validate(2)
        with self.assertRaises(ValidationError):
            validator.validate(2.0)

        with self.assertRaises(ValidationError):
            validator.validate(2.1)

        validator = Draft04Validator(type='integer', maximum=2, exclusiveMaximum=True)
        with self.assertRaises(ValidationError):
            validator.validate(2)
        with self.assertRaises(ValidationError):
            validator.validate(2.0)
        with self.assertRaises(ValidationError):
            validator.validate(2.1)
        with self.assertRaises(ValidationError):
            validator.validate(1.9991)

    def test_multiple(self):
        validator = Draft04Validator(type='integer', multipleOf=2)
        validator.validate(2)

        with self.assertRaises(ValidationError):
            validator.validate(2.0)

        with self.assertRaises(ValidationError):
            validator.validate(3)

        validator = Draft04Validator(type='integer', multipleOf=2.5)
        with self.assertRaises(ValidationError):
            validator.validate(2.5)
        validator.validate(50)
        validator.validate(-50)

        with self.assertRaises(ValidationError):
            validator.validate(-51)


class TestString(TestCase):

    def test_any_string(self):
        validator = Draft04Validator(type='string')
        validator.validate('foo')
        validator.validate(u'bar')

        with self.assertRaises(ValidationError):
            validator.validate(1)

        with self.assertRaises(ValidationError):
            validator.validate(True)

    def test_length(self):
        validator = Draft04Validator(type='string', minLength=3, maxLength=10)
        validator.validate('abc')
        validator.validate('abcdefghij')
        with self.assertRaises(ValidationError):
            validator.validate('a')
        with self.assertRaises(ValidationError):
            validator.validate('abcdefghijk')

    def test_pattern(self):
        validator = Draft04Validator(type='string', pattern='''foo|bar''')
        validator.validate('foo')
        with self.assertRaises(ValidationError):
            validator.validate('baz')


class TestArray(TestCase):
    def test_any_array(self):
        validator = Draft04Validator(type='array')
        validator.validate([])
        validator.validate(['foo'])
        with self.assertRaises(ValidationError):
            validator.validate('foo')
        with self.assertRaises(ValidationError):
            validator.validate(123)
        with self.assertRaises(ValidationError):
            validator.validate({})

    def test_items(self):
        validator = Draft04Validator(type='array',
                              items=[{}, {}, {}],
                              additionalItems=False)
        validator.validate([])
        validator.validate([[1, 2, 3, 4], [5, 6, 7, 8]])
        validator.validate([1, 2, 3])
        with self.assertRaises(ValidationError):
            validator.validate([1, 2, 3, 4])
        with self.assertRaises(ValidationError):
            validator.validate([None, {'a': 'b'}, True, 31.000002020013])


class TestObject(TestCase):
    def test_any_object(self):
        validator = Draft04Validator(type='object')
        validator.validate({})
        validator.validate({'foo': 'bar'})
        with self.assertRaises(ValidationError):
            validator.validate('foo')
        with self.assertRaises(ValidationError):
            validator.validate(123)
        with self.assertRaises(ValidationError):
            validator.validate([])


class TestGeneral(TestCase):
    def test_enum(self):
        validator = Draft04Validator(enum=['foo', 42])
        validator.validate('foo')
        validator.validate(42)
        with self.assertRaises(ValidationError):
            validator.validate('bar')

    def test_format(self):
        pass


class TestCollections(TestCase):
    def test_all_of(self):
        foo = Draft04Validator(type='string', enum=['foo'])
        bar = Draft04Validator(type='string', pattern='^f[o]+$')
        validator = Draft04Validator(type='string', allOf=[foo, bar])
        validator.validate('foo')
        with self.assertRaises(ValidationError):
            validator.validate('bar')

    def test_any_of(self):
        foo = Draft04Validator(type='string', enum=['foo'])
        bar = Draft04Validator(type='string', enum=['bar'])
        validator = Draft04Validator(type='string', anyOf=[foo, bar])
        validator.validate('foo')
        validator.validate('bar')
        with self.assertRaises(ValidationError):
            validator.validate('baz')

    def test_one_of(self):
        foo = Draft04Validator(type='string', enum=['foo'])
        bar = Draft04Validator(type='string', pattern='^f[o]+$')
        baz = Draft04Validator(type='string', enum=['bar'])
        validator = Draft04Validator(type='string', oneOf=[foo, bar, baz])
        validator.validate('bar')
        with self.assertRaises(ValidationError):
            validator.validate('foo')



"""
    json.pointer
    ~~~~~~~~~~~~

    JSON Pointer defines a string syntax for identifying a specific value
    within a JavaScript Object Notation (JSON) document.

"""

from __future__ import absolute_import, print_function, unicode_literals

__all__ = ['extract', 'DocumentPointer',
           'Pointer', 'PointerToken', 'ExtractError']

import logging
from .bases import DocumentPointer, Pointer, PointerToken
from .exceptions import ExtractError

logger = logging.getLogger(__name__)


def extract(obj, pointer, bypass_ref=False):
    """Extract member or attributes from a pointer"""
    return Pointer(pointer).extract(obj, bypass_ref)




"""
    tests.tests_errors
    ~~~~~~~~~~~~~~~~~~

"""


from json.schema import load
from json.schema.exceptions import ValidationError
from . import TestCase, fixture


class TestErrors(TestCase):

    def test_check(self):
        validator = load(fixture('five.schema.json'))

        try:
            validator.validate({
                'creditcard': {
                    'provider': 'visa'
                }
            })
        except ValidationError as error:
            pass
        else:
            self.fail("shouldn't happen")

        try:
            validator.validate({
                'creditcard': {
                    'provider': 'mastercard',
                    'securitycode': 123
                }
            })
        except ValidationError as error:
            pass
        else:
            self.fail("shouldn't happen")



"""
    tests.tests_reference
    ~~~~~~~~~~~~~~~~~~~~~

"""

from json.pointer import extract, ExtractError, DocumentPointer, Pointer
from . import TestCase


class TestPointer(TestCase):
    document = {
        'foo': ['bar', 'baz', {
            '$ref': 'obj2#/sub'
        }]
    }

    def test_simple(self):
        assert 'baz' == extract(self.document, '/foo/1')

    def test_with_reference(self):
        with self.assertRaises(ExtractError):
            assert 'quux' == extract(self.document, '/foo/2')

    def test_bypass_reference(self):
        assert 'obj2#/sub' == extract(self.document, '/foo/2/$ref',
                                      bypass_ref=True)

    def test_compare(self):
        assert '/foo' == Pointer('/foo')
        assert Pointer('/foo') != Pointer('/bar')
        tokens = Pointer('//a~1b/c%d/e^f/g|h/i\\j/k\"l/ /m~0n').tokens
        assert tokens == ['', 'a/b', 'c%d', 'e^f', 'g|h', 'i\\j', 'k\"l', ' ', 'm~n']  # noqa

    def test_iteration(self):
        obj = self.document
        for token in Pointer('/foo/1'):
            obj = token.extract(obj)
        assert 'baz' == obj

    def test_document(self):
        dp = DocumentPointer('example.com#/foo')
        assert not dp.is_inner()
        assert dp.document == 'example.com'
        assert dp.pointer == '/foo'

    def test_inner_document(self):
        dp = DocumentPointer('#/foo')
        assert dp.is_inner()
        assert dp.document == ''
        assert dp.pointer == '/foo'



"""
    json.schema.exceptions
    ~~~~~~~~~~~~~~~~~~~~~~

"""

__all__ = ['CompilationError', 'ReferenceError', 'ValidationError']


class CompilationError(Exception):
    """Raised while schema parsing"""
    def __init__(self, message, schema):
        super(CompilationError, self).__init__(message, schema)
        self.schema = schema


class ReferenceError(Exception):
    """Raised while reference error"""
    def __init__(self, *args):
        super(ReferenceError, self).__init__(*args)


class ValidationError(Exception):
    """Raised while validating an obj with JsonSchema"""
    def __init__(self, message, obj=None, rule=None):
        super(ValidationError, self).__init__(message)
        self.obj = obj
        self.rule = rule



"""
    json.pointer.exceptions
    ~~~~~~~~~~~~~~~~~~~~~~~

"""

__all__ = ['ExtractError']


class ExtractError(Exception):
    def __init__(self, obj, *args):
        super(ExtractError, self).__init__(*args)
        self.obj = obj



"""
    json.schema.util
    ~~~~~~~~~~~~~~~~

"""

__all__ = ['rfc3339_to_datetime']

from datetime import tzinfo, timedelta, datetime, date
import time


class offset(tzinfo):
    def __init__(self, value):
        self.value = value
        # super(tzinfo, self).__init__()

    def utcoffset(self, dt):
        hours, minutes = self.value.split(':', 1)
        return timedelta(hours=int(hours), minutes=int(minutes))

    def tzname(self, dt):
        return '{}'.format(self.value)


def rfc3339_to_datetime(data):
    """convert a rfc3339 date representation into a Python datetime"""
    try:
        ts = time.strptime(data, '%Y-%m-%d')
        return date(*ts[:3])
    except ValueError as error:
        pass

    try:
        dt, _, tz = data.partition('Z')
        if tz:
            tz = offset(tz)
        else:
            tz = offset('00:00')
        ts = time.strptime(dt, '%Y-%m-%dT%H:%M:%S.%f')
        return datetime(*ts[:6], tzinfo=tz)
    except ValueError:
        raise ValueError('date-time {!r} is not a valid rfc3339 date representation'.format(data))  # noqa



"""
    json.reference.bases
    ~~~~~~~~~~~~~~~~~~~~

"""

__all__ = ['LocalRegistry', 'Registry']

import logging
from .exceptions import NotFound, Forbidden
from .util import ref, MutableMapping
from json.pointer import DocumentPointer

logger = logging.getLogger(__name__)


class Registry(MutableMapping):
    """Register all documents.

    :ivar docs: all documents

    """

    def __init__(self, docs=None):
        self.docs = docs or {}
        super(Registry, self).__init__()

    def prototype(self, dp):
        obj = self[dp.document]
        return self[dp.document], LocalRegistry(obj, self)

    def resolve(self, pointer):
        """Resolve from documents.

        :param pointer: foo
        :type pointer: DocumentPointer
        """

        dp = DocumentPointer(pointer)
        obj, fetcher = self.prototype(dp)

        for token in dp.pointer:
            obj = token.extract(obj, bypass_ref=True)
            reference = ref(obj)
            if reference:
                obj = fetcher.resolve(reference)
        return obj

    def __getitem__(self, uri):
        try:
            return self.docs[uri]
        except KeyError:
            return NotFound('{!r} not registered'.format(uri))

    def __setitem__(self, uri, obj):
        self.docs[uri] = obj

    def __delitem__(self, uri):
        del self.docs[uri]

    def __len__(self):
        return len(self.docs)

    def __iter__(self):
        return iter(self.docs)


class LocalRegistry(Registry):
    """Scoped registry to a local document.

    :ivar doc: the local document
    :ivar registry: all documents
    :ivar key: current document identifier

    """

    key = '<local>'

    def __init__(self, doc, registry=None):
        self.doc = doc
        self.registry = registry or {}

    def prototype(self, dp):
        if dp.is_inner():
            return self.doc, self
        else:
            obj = self[dp.document]
            return self[dp.document], LocalRegistry(obj, self)

    def __getitem__(self, uri):
        try:
            return self.doc if uri == self.key else self.registry[uri]
        except (NotFound, KeyError):
            return NotFound('{!r} not registered'.format(uri))

    def __setitem__(self, uri, obj):
        if uri == self.key:
            raise Forbidden('setting {} is forbidden'.format(self.key))
        self.registry[uri] = obj

    def __delitem__(self, uri):
        if uri == self.key:
            raise Forbidden('deleting {} is forbidden'.format(self.key))
        del self.registry[uri]

    def __len__(self):
        return len(set(list(self.registry.keys()) + [self.key]))

    def __iter__(self):
        yield self.key
        for key in self.registry.keys():
            yield key



"""
    json.schema.draft04
    ~~~~~~~~~~~~~~~~~~~

    Implementation of JSON Schema draft04.

"""

from __future__ import absolute_import, print_function, unicode_literals

__all__ = ['compile', 'Draft04Validator']

from copy import deepcopy
import itertools
import logging
import os.path
import re

from six import integer_types, string_types, binary_type, PY2

from .exceptions import CompilationError, ValidationError
from .bases import Validator, ReferenceValidator
from .util import rfc3339_to_datetime

logger = logging.getLogger(__name__)


def absolute(src, dest):
    if not dest.startswith('#'):
        return dest

    a, _, b = src.partition('#')
    c, _, d = dest.partition('#')
    if not c:
        return '{}#{}'.format(a, d)
    return dest


def compile(schema, uri, factory):
    """
    Compile python object into a Draft04Validator instance.
    """

    if '$ref' in schema:
        return ReferenceValidator(absolute(uri, schema['$ref']),
                                  factory,
                                  'http://json-schema.org/draft-04/schema#')

    schema = deepcopy(schema)
    attrs = {}
    attrs['title'] = schema.get('title', None)
    attrs['description'] = schema.get('description', None)
    if 'default' in schema:
        attrs['default'] = schema.get('default')
    attrs['uri'] = uri
    attrs['type'] = schema.get('type', None)

    if 'enum' in schema:
        attr = schema['enum']
        if not isinstance(attr, list):
            raise CompilationError('enum must be a list', schema)
        attrs['enum'] = attr

    if 'not' in schema:
        attr = schema['not']
        if not isinstance(attr, dict):
            raise CompilationError('not must be a dict', schema)
        attrs['not'] = compile(attr, os.path.join(uri, 'not'), factory)
    for name in ('allOf', 'anyOf', 'oneOf'):
        if name in schema:
            attr = schema[name]
            if not isinstance(attr, list):
                raise CompilationError('{} must be a list'.format(name),
                                       schema)
            sub_uri = os.path.join(uri, name)
            for i, element in enumerate(attr):
                attr[i] = compile(element, sub_uri, factory)
            attrs[name] = attr

    for name in ('items', 'additionalItems'):
        if name not in schema:
            continue

        attr = schema[name]
        sub_uri = os.path.join(uri, name)
        if isinstance(attr, list):
            # each value must be a json schema
            attr = [compile(attr, sub_uri, factory) for element in attr]
        elif isinstance(attr, dict):
            # value must be a json schema
            attr = compile(attr, sub_uri, factory)
        elif not isinstance(attr, bool):
            # should be a boolean
            raise CompilationError('wrong type for {}'.format(name), schema)  # noqa

        attrs[name] = attr

    for name in ('maxItems', 'minItems'):
        if name not in schema:
            continue
        attr = schema[name]
        if not isinstance(attr, int):
            # should be a boolean
            raise CompilationError('{} must be a integer'.format(name), schema)  # noqa
        if attr < 0:
            # should be a boolean
            raise CompilationError('{} must be greater than 0'.format(name), schema)  # noqa
        attrs[name] = attr

    if 'uniqueItems' in schema:
        attr = schema['uniqueItems']
        if not isinstance(attr, bool):
            raise CompilationError('{} must be a bool'.format(name), schema)  # noqa
        attrs[name] = attr

    for name in ('maxProperties', 'minProperties'):
        if name in schema:
            attr = schema[name]
            attrs[name] = attr

    for name in ('properties', 'patternProperties'):
        if name in schema:
            if not isinstance(schema[name], dict):
                raise CompilationError('{} must be a dict'.format(name),
                                       schema)
            attr = {}
            for subname, subschema in schema[name].items():
                attr[subname] = compile(subschema, os.path.join(uri, name),
                                        factory)
            attrs[name] = attr

    for name in ('maxProperties', 'minProperties'):
        if name in schema:
            attr = schema[name]
            if not isinstance(attr, integer_types):
                raise CompilationError('{} must be an integer'.format(name),
                                       schema)
            if attr < 0:
                raise CompilationError('{} must be greater than 0'.format(name),  # noqa
                                       schema)
            attrs[name] = attr

    if 'additionalProperties' in schema:
        attr = schema['additionalProperties']
        if isinstance(attr, dict):
            attr = compile(attr, os.path.join(uri, name), factory)
        elif attr is True:
            attr = compile({}, os.path.join(uri, name), factory)
        elif not isinstance(schema['additionalProperties'], bool):
            raise CompilationError('additionalProperties must be '
                                   'a dict or a bool', schema)
        attrs['additionalProperties'] = attr

    if 'required' in schema:
        attrs['required'] = schema['required'][:]

    if 'multipleOf' in schema:
        attr = schema['multipleOf']
        if not isinstance(attr, (int, float)):
            raise CompilationError('multipleOf must be a number', schema)
        if attr < 0:
            raise CompilationError('multipleOf must be greater than 0',
                                   schema)
        attrs['multipleOf'] = attr

    if 'minimum' in schema:
        attr = schema['minimum']
        if not isinstance(attr, (int, float)):
            raise CompilationError('minimum must be a number', schema)
        attrs['minimum'] = attr
    if 'exclusiveMinimum' in schema:
        attr = schema['exclusiveMinimum']
        if not isinstance(attr, bool):
            raise CompilationError('exclusiveMinimum must be a boolean',
                                   schema)
        if 'minimum' not in schema:
            raise CompilationError('exclusiveMinimum reclame maximum',
                                   schema)
        attrs['exclusiveMinimum'] = attr

    if 'maximum' in schema:
        attr = schema['maximum']
        if not isinstance(attr, (int, float)):
            raise CompilationError('maximum must be a number', schema)
        attrs['maximum'] = attr
    if 'exclusiveMaximum' in schema:
        attr = schema['exclusiveMaximum']
        if not isinstance(attr, bool):
            raise CompilationError('exclusiveMaximum must be a boolean',
                                   schema)
        if 'maximum' not in schema:
            raise CompilationError('exclusiveMaximum reclame maximum',
                                   schema)
        attrs['exclusiveMaximum'] = attr

    if 'maxLength' in schema:
        attr = 'maxLength'
        if not isinstance(attr, int):
            raise CompilationError('maxLength must be an integer', schema)
        if attr < 0:
            raise CompilationError('maxLength must be equal '
                                   'or greater than 0', schema)
        attrs['maxLength'] = attr

    if 'minLength' in schema:
        attr = 'minLength'
        if not isinstance(attr, int):
            raise CompilationError('minLength must be an integer', schema)
        if attr < 0:
            raise CompilationError('minLength must be equal '
                                   'or greater than 0', schema)
        attrs['minLength'] = attr

    for name in ('pattern', 'format'):
        if name in schema:
            attr = schema[name]
            if not isinstance(attr, string_types):
                raise CompilationError('{} must be a string'.format(name), schema)
            attrs[name] = attr

    return Draft04Validator(**attrs)


class Items(object):
    """
    Describes items and additionalItems rules
    """

    def __init__(self, key):
        self.key = key

    def __get__(self, obj, type=None):
        iterable = getattr(obj, self.key, [])
        if isinstance(iterable, Validator):
            return itertools.cycle([iterable])
        if not iterable:
            return []
        return iterable

    def __set__(self, obj, value):
        setattr(obj, self.key, value)


class Draft04Validator(Validator):
    """
    Implements a draft04 validator.
    """

    items = Items('_items')
    additionalItems = Items('_additionalItems')

    def __init__(self, **attrs):
        self.id = attrs.pop('id', None)
        if self.id:
            logger.info('id is not implemented and is here for information purpose')
        self.uri = attrs.pop('uri', None)
        self.title = attrs.pop('title', None)
        self.description = attrs.pop('description', None)
        self.type = attrs.pop('type', None)
        self.default = attrs.pop('default', None)
        self.enum = attrs.pop('enum', [])
        self.negate = attrs.pop('not', None)
        self.allOf = attrs.pop('allOf', [])
        self.anyOf = attrs.pop('anyOf', [])
        self.oneOf = attrs.pop('oneOf', [])

        self.initialize_array(attrs)
        self.initialize_object(attrs)
        self.initialize_number(attrs)
        self.initialize_string(attrs)

    def initialize_array(self, attrs):
        self.items = attrs.pop('items', {})
        self.additionalItems = attrs.pop('additionalItems', False)
        self.maxItems = attrs.pop('maxItems', None)
        self.minItems = attrs.pop('minItems', 0)
        self.uniqueItems = attrs.pop('uniqueItems', None)

    def initialize_object(self, attrs):
        self.properties = attrs.pop('properties', {})
        self.patternProperties = attrs.pop('patternProperties', {})
        self.additionalProperties = attrs.pop('additionalProperties', {})
        self.maxProperties = attrs.pop('maxProperties', None)
        self.minProperties = attrs.pop('minProperties', None)
        self.required = attrs.pop('required', [])

    def initialize_number(self, attrs):
        self.minimum = attrs.pop('minimum', None)
        self.maximum = attrs.pop('maximum', None)
        self.exclusiveMinimum = attrs.pop('exclusiveMinimum', False)
        self.exclusiveMaximum = attrs.pop('exclusiveMaximum', False)
        self.multipleOf = attrs.pop('multipleOf', None)

    def initialize_string(self, attrs):
        self.maxLength = attrs.pop('maxLength', None)
        self.minLength = attrs.pop('minLength', None)
        self.pattern = attrs.pop('pattern', None)
        self.format = attrs.pop('format', None)

    def has_default(self):
        return getattr(self, 'default', False)

    def has_type(self, type):
        return self.type and (self.type == type or type in self.type)

    def validate(self, obj):
        """
        Validates obj.

        Note that the returned object is not the same as object.
        It may contains default values.

        :param obj: The object to validate
        :return: the validated obj
        :raises ValidationError: Whole or part of obj does not validate.
        """

        # common
        self.validate_enum(obj)
        self.validate_negate(obj)
        self.validate_and(obj)
        self.validate_or(obj)
        self.validate_xor(obj)

        # primitives
        if self.validate_array(obj):
            obj = self.validate_items(obj)

        if self.validate_boolean(obj):
            pass

        if self.validate_integer(obj):
            self.validate_minimum(obj)
            self.validate_maximum(obj)
            self.validate_multiple(obj)

        if self.validate_null(obj):
            pass

        if self.validate_number(obj):
            self.validate_minimum(obj)
            self.validate_maximum(obj)
            self.validate_multiple(obj)

        if self.validate_object(obj):
            obj = self.validate_properties(obj)

        if self.validate_string(obj):
            self.validate_length(obj)
            self.validate_pattern(obj)
            self.validate_format(obj)

        return obj

    def validate_enum(self, obj):
        if self.enum and obj not in self.enum:
            raise ValidationError('obj is not allowed', obj, rule=self.uri)

    def validate_array(self, obj):
        if isinstance(obj, list):
            return True
        elif self.has_type('array'):
            raise ValidationError('obj must be an array', obj, rule=self.uri)
        return False

    def validate_boolean(self, obj):
        if obj in (True, False):
            return True
        elif self.has_type('boolean'):
            raise ValidationError('obj must be a boolean', obj, rule=self.uri)
        return False

    def validate_integer(self, obj):
        if isinstance(obj, integer_types) and not isinstance(obj, bool):
            return True
        elif self.has_type('integer'):
            raise ValidationError('obj must be an integer', obj, rule=self.uri)
        return False

    def validate_null(self, obj):
        if obj is None:
            return True
        elif self.has_type('null'):
            raise ValidationError('obj must be null', obj, rule=self.uri)
        return False

    def validate_number(self, obj):
        if isinstance(obj, (integer_types, float)) and not isinstance(obj, bool):  # noqa
            return True
        elif self.has_type('number'):
            raise ValidationError('obj must be a number', obj, rule=self.uri)
        return False

    def validate_object(self, obj):
        if isinstance(obj, dict):
            return True
        elif self.has_type('object'):
            raise ValidationError('obj must be an object', obj, rule=self.uri)
        return False

    def validate_string(self, obj):
        if isinstance(obj, string_types):
            return True
        if PY2 and isinstance(obj, binary_type):
            return True
        elif self.has_type('string'):
            raise ValidationError('obj must be a string', obj, rule=self.uri)
        return False

    def validate_length(self, obj):
        l = len(obj)
        if self.minLength and l < self.minLength:
            raise ValidationError('length of obj must be greater or equal'
                                  ' than {}'.format(self.minLength), obj, rule=self.uri)
        if self.maxLength and l > self.maxLength:
            raise ValidationError('length of obj must be lesser or equal'
                                  ' than {}'.format(self.minLength), obj, rule=self.uri)

    def validate_pattern(self, obj):
        if self.pattern and not self.regex.match(obj):
            raise ValidationError('obj does not validate '
                                  '{!r} pattern'.format(self.pattern), obj, rule=self.uri)

    def validate_format(self, obj):
        if not self.format:
            return obj
        if self.format == 'date-time':
            return self.validate_datetime(obj)
        if self.format == 'email':
            return self.validate_email(obj)
        if self.format == 'hostname':
            return self.validate_hostname(obj)
        if self.format == 'ipv4':
            return self.validate_ipv4(obj)
        if self.format == 'ipv6':
            return self.validate_ipv6(obj)
        if self.format == 'uri':
            return self.validate_uri(obj)
        raise ValidationError('format {} is not '
                              'defined.'.format(self.pattern), obj, rule=self.uri)

    @property
    def regex(self):
        if not hasattr(self, '_regex'):
            setattr(self, '_regex', re.compile(self.pattern))
        return self._regex

    def validate_minimum(self, obj):
        if self.minimum is not None:
            if not obj >= self.minimum:
                raise ValidationError('object must be greater '
                                      'than {}'.format(self.minimum), obj, rule=self.uri)
            if self.exclusiveMinimum and obj == self.minimum:
                raise ValidationError('object must be greater '
                                      'than {}'.format(self.minimum), obj, rule=self.uri)

    def validate_maximum(self, obj):
        if self.maximum is not None:
            if not obj <= self.maximum:
                raise ValidationError('object must be lesser '
                                      'than {}'.format(self.maximum), obj, rule=self.uri)
            if self.exclusiveMaximum and obj == self.maximum:
                raise ValidationError('object must be lesser '
                                      'than {}'.format(self.maximum), obj, rule=self.uri)

    def validate_multiple(self, obj):
        if self.multipleOf and not obj % self.multipleOf == 0:
            raise ValidationError('object must be a multiple '
                                  'of {}'.format(self.multipleOf), obj, rule=self.uri)

    def validate_items(self, obj):
        l = len(obj)
        if self.minItems and l < self.minItems:
            raise ValidationError('object must have at least '
                                  '{} elements'.format(self.minItems), obj, rule=self.uri)
        if self.maxItems and l > self.maxItems:
            raise ValidationError('object must have less than '
                                  '{} elements'.format(self.maxItems), obj, rule=self.uri)

        if self._items == {}:
            # validation of the instance always succeeds
            # regardless of the value of "additionalItems"
            return obj
        if self._additionalItems in (True, {}):
            # validation of the instance always succeeds
            return obj

        schemas = itertools.chain(self.items, self.additionalItems)
        iteration = 0
        errors = []
        for sub, schema in zip(obj, schemas):
            iteration += 1
            try:
                if schema == {}:
                    continue
                else:
                    schema.validate(sub)
            except ValidationError as error:
                errors.append(error)
        if errors:
            raise ValidationError(errors)

        if iteration < len(obj):
            raise ValidationError('object has too much elements', obj, rule=self.uri)

        if self.uniqueItems and len(set(obj)) != len(obj):
            raise ValidationError('items must be unique', obj, rule=self.uri)

        return obj

    def validate_properties(self, obj):
        l = len(obj)
        if isinstance(self.maxProperties, integer_types) and l > self.maxProperties:  # noqa
            raise ValidationError('too much properties, '
                                  'max {}'.format(self.maxProperties), obj, rule=self.uri)
        if isinstance(self.minProperties, integer_types) and l < self.minProperties:  # noqa
            raise ValidationError('too few properties, '
                                  'min {}'.format(self.minProperties), obj, rule=self.uri)

        obj = deepcopy(obj)
        errors, missing = [], set(obj.keys())
        missing.update(self.required)
        missing.update(self.properties.keys())
        for member, schema in self.properties.items():
            if member in obj:
                try:
                    schema.validate(obj[member])
                    missing.discard(member)
                except ValidationError as error:
                    errors.append(error)
                except AttributeError:
                    raise
            elif schema.has_default():
                obj[member] = deepcopy(schema.default)
        if errors:
            raise ValidationError(errors)
        for pattern, schema in self.patternProperties.items():
            regex = re.compile(pattern)
            for member, value in obj.items():
                if regex.match(member):
                    try:
                        schema.validate(value)
                        missing.discard(member)
                    except ValidationError as error:
                        errors.append(error)
        if errors:
            raise ValidationError(errors)

        if missing:
            if self.additionalProperties in ({}, True):
                missing.clear()
            elif self.additionalProperties:
                schema = self.additionalProperties
                for member in missing:
                    try:
                        obj[member] = schema.validate(obj[member])
                        missing.discard(member)
                    except ValidationError as error:
                        errors.append(error)
        for member in self.required:
            if member not in obj:
                missing.add(member)
        if errors:
            raise ValidationError(errors, obj, rule=self.uri)

        if missing:
            raise ValidationError('missing definitions for {}'.format(missing), obj, rule=self.uri)

        return obj

    def validate_negate(self, obj):
        if self.negate:
            try:
                self.negate.validate(obj)
            except ValidationError:
                pass
            else:
                raise ValidationError('obj is not allowed', obj, rule=self.uri)

    def validate_and(self, obj):
        if self.allOf:
            errors = []
            for validator in self.allOf:
                try:
                    validator.validate(obj)
                except ValidationError as error:
                    errors.append(error)
            if errors:
                raise ValidationError(errors)
        return obj

    def validate_or(self, obj):
        if self.anyOf:
            errors = []
            for validator in self.anyOf:
                try:
                    return validator.validate(obj)
                except ValidationError as error:
                    errors.append(error)
            if errors:
                raise ValidationError(errors)
        return obj

    def validate_xor(self, obj):
        if self.oneOf:
            ok, errors = [], []
            for validator in self.oneOf:
                try:
                    ok.append(validator.validate(obj))
                except ValidationError as error:
                    errors.append(error)
            if len(ok) == 1:
                return ok.pop()
            if errors:
                raise ValidationError(errors)
        return obj

    def validate_datetime(self, obj):
        try:
            return rfc3339_to_datetime(obj)
        except ValueError:
            raise ValidationError('{!r} is not a valid datetime'.format(obj), obj, rule=self.uri)

    def validate_email(self, obj):
        pattern = re.compile('[^@]+@[^@]+\.[^@]+')
        if not pattern.match(obj):
            raise ValidationError('{!r} is not defined'.format(obj), obj, rule=self.uri)

    def validate_hostname(self, obj):
        try:
            host = deepcopy(obj)
            if len(host) > 255:
                raise ValueError
            if host[-1] == ".":
                host = host[:-1] # strip exactly one dot from the right, if present
            allowed = re.compile("(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
            if not all(allowed.match(x) for x in host.split(".")):
                raise ValueError
            return obj
        except ValueError:
            raise ValidationError('{!r} is not a valid hostname'.format(obj), obj, rule=self.uri)

    def validate_ipv4(self, obj):
        try:
            parts = obj.split('.', 3)
            for part in parts:
                part = int(part)
                if part > 127 or part < 0:
                    raise ValueError
        except ValueError:
            raise ValidationError('{!r} is not an ipv4'.format(obj), obj, rule=self.uri)

    def validate_ipv6(self, obj):
        raise ValidationError('{!r} is not defined'.format(obj), obj, rule=self.uri)

    def validate_uri(self, obj):
        raise ValidationError('{!r} is not defined'.format(obj), obj, rule=self.uri)



"""
    tests.tests_util
    ~~~~~~~~~~~~~~~~

"""

from json.schema.util import rfc3339_to_datetime
from . import TestCase


class TestDates(TestCase):
    def test_first(self):
        res = rfc3339_to_datetime('1985-04-12T23:20:50.52')
        assert res.hour == 23, res.hour

        res = rfc3339_to_datetime('1985-04-12T23:20:50.52')
        assert res.hour == 23

        with self.assertRaises(ValueError):
            res = rfc3339_to_datetime('foobar')
            assert res.hour == 23



"""
    json.reference.util
    ~~~~~~~~~~~~~~~~~~~

"""

__all__ = ['ref', 'MutableMapping']


def ref(obj):
    """Extracts $ref of object."""
    try:
        return obj['$ref']
    except (KeyError, TypeError):
        return None

try:
    # py3
    from collections.abc import MutableMapping
except ImportError:
    # py2
    from collections import MutableMapping



"""
    json.pointer.bases
    ~~~~~~~~~~~~~~~~~~

"""


__all__ = ['DocumentPointer', 'Pointer', 'PointerToken']

import logging
from six import string_types
from .exceptions import ExtractError

logger = logging.getLogger(__name__)


class DocumentPointer(object):
    """Defines a document pointer

    :ivar document: document name
    :ivar pointer: pointer
    """

    def __init__(self, pointer):
        """
        :param pointer: a string or DocumentPointer instance
        """
        if isinstance(pointer, DocumentPointer):
            document, path = pointer
        elif '#' not in pointer:
            raise ValueError('# is missing')
        else:
            document, path = pointer.split('#', 1)
        self.document = document
        self.pointer = Pointer(path)

    def extract(self, obj, bypass_ref=False):
        """
        Extract subelement from obj, according to pointer.
        It assums that document is the object.

        :param obj: the object source
        :param bypass_ref: disable JSON Reference errors
        """
        return self.pointer.extract(obj, bypass_ref)

    def is_inner(self):
        """Tells if pointer refers to an inner document
        """
        return self.document == ''

    def __iter__(self):
        return iter([self.document, self.pointer])

    def __eq__(self, other):
        if isinstance(other, string_types):
            return other == self.__str__()
        return super(Pointer, self).__eq__(other)

    def __str__(self):
        return '{}#{}'.format(self.document, self.pointer)


class Pointer(object):
    """Defines a pointer

    :ivar tokens: list of PointerToken
    """

    def __init__(self, pointer):
        """
        :param pointer: a string or Pointer instance
        """
        if isinstance(pointer, Pointer):
            tokens = pointer.tokens[:]
        elif pointer == '':
            tokens = []
        elif not pointer.startswith('/'):
            raise ValueError('pointer must start with /', pointer)
        else:
            # decode now
            tokens = []
            for part in pointer[1:].split('/'):
                part = part.replace('~1', '/')
                part = part.replace('~0', '~')
                tokens.append(PointerToken(part))
        self.tokens = tokens

    def extract(self, obj, bypass_ref=False):
        """
        Extract subelement from obj, according to tokens.

        :param obj: the object source
        :param bypass_ref: disable JSON Reference errors
        """
        for token in self.tokens:
            obj = token.extract(obj, bypass_ref)
        return obj

    def __iter__(self):
        return iter(self.tokens)

    def __eq__(self, other):
        if isinstance(other, string_types):
            return other == self.__str__()
        return super(Pointer, self).__eq__(other)

    def __str__(self):
        output = ''
        for part in self.tokens:
            part = part.replace('~', '~0')
            part = part.replace('/', '~1')
            output += '/' + part
        return output

    def __repr__(self):
        return '<{}({!r})>'.format(self.__class__.__name__, self.__str__())


class PointerToken(str):
    """
    A single token
    """

    def extract(self, obj, bypass_ref=False):
        """
        Extract subelement from obj, according to current token.

        :param obj: the object source
        :param bypass_ref: disable JSON Reference errors
        """
        try:
            if isinstance(obj, dict):
                if not bypass_ref and '$ref' in obj:
                    raise ValueError('presence of a $ref member')
                obj = self.extract_mapping(obj)
            elif isinstance(obj, (list, tuple)):
                obj = self.extract_sequence(obj)
            else:
                raise TypeError('not for this kind', self, obj)

            if isinstance(obj, dict):
                if not bypass_ref and '$ref' in obj:
                    raise ValueError('presence of a $ref member', bypass_ref)
            return obj
        except Exception as error:
            logger.exception(error)
            args = [arg for arg in error.args if arg not in (self, obj)]
            raise ExtractError(obj, *args)

    def extract_mapping(self, obj):
        if self in obj:
            return obj[self]

        if self.isdigit():
            key = int(self)
            if key in obj:
                return obj[key]

        raise KeyError('member not found', self, obj)

    def extract_sequence(self, obj):
        if self == '-':
            raise ValueError('refers to a nonexistent array element')
        if not self.isdigit():
            raise TypeError('not for this kind', self, obj)
        try:
            return obj[int(self)]
        except IndexError:
            raise IndexError('element not found', self, obj)

    def __repr__(self):
        return '<{}({!r})>'.format(self.__class__.__name__, str(self))



#!/usr/bin/env python

# http://alexis.notmyidea.org/distutils2/setupcfg.html

from setuptools import setup, find_packages

setup(
    setup_requires=['pbr'],
    pbr=True,
    namespace_packages=['json'],
    packages=find_packages('src'),
    package_dir = {'': 'src'},
)


"""
    tests.tests_reference
    ~~~~~~~~~~~~~~~~~~~~~

"""

from json.reference import resolve, Registry
from . import TestCase


class TestReference(TestCase):
    def test_local(self):
        obj = {
            'foo': ['bar', 'baz']
        }

        assert 'baz' == resolve(obj, '#/foo/1')

    def test_inner_reference(self):
        obj = {
            'foo': ['bar', 'baz', {
                '$ref': '#/sub'
            }],
            'sub': 'quux'
        }

        assert 'quux' == resolve(obj, '#/foo/2')

    def test_outside_reference(self):
        obj = {
            'foo': ['bar', 'baz', {
                '$ref': 'obj2#/sub'
            }]
        }
        registry = {
            'obj2': {'sub': 'quux'}
        }

        assert 'quux' == resolve(obj, '#/foo/2', registry=registry)

    def test_outside_reference_2(self):
        registry = Registry({
            'obj1': {
                'foo': ['bar', 'baz', {
                    '$ref': 'obj2#/sub/0'
                }]
            },
            'obj2': {
                'sub': [
                    {'$ref': '#/sub2'}
                ],
                'sub2': 'quux'
            }
        })
        assert 'quux' == registry.resolve('obj2#/sub/0')
        # assert 'quux' == resolve(obj, '#/foo/2', registry=registry)



"""
    json.schema
    ~~~~~~~~~~~

    JSON Schema defines the media type "application/schema+json",
    a JSON based format for defining the structure of JSON data.

    JSON Schema provides a contract for what JSON data is required
    for a given application and how to interact with it.

    JSON Schema is intended to define validation, documentation,
    hyperlink navigation, and interaction control of JSON data.

"""

from __future__ import absolute_import, print_function, unicode_literals

__all__ = ['load', 'load_from_file', 'factory']

from copy import deepcopy
from json import load as file_load

from .exceptions import CompilationError
from .draft04 import compile, Draft04Validator
from .bases import Validator, ReferenceValidator


def load(schema, **kwargs):
    """load schema from a python object

    :param schema: The schema
    :type schema: dict
    :return: The validator instance
    :rtype: Validator
    :raises CompilationError: Whole or part of schema failed to be loaded.
    """
    return factory(schema, '<document>#', **kwargs)


def load_from_file(filename, **kwargs):
    """load schema from a file

    :param filename: the json file to be loaded
    :type filename: str
    :return: The validator instance
    :rtype: Validator
    :raises CompilationError: Whole or part of schema failed to be loaded.
    """
    with open(filename, 'r') as file:
        schema = file_load(file)
    return factory(schema, '<document>#', **kwargs)


def factory(schema, uri, default_spec=None, loader=None):

    instance = Factory(loader)
    return instance(schema, uri)

    default_spec = default_spec or 'http://json-schema.org/draft-04/schema#'
    spec = schema.get('$schema', default_spec)
    if spec != 'http://json-schema.org/draft-04/schema#':
        raise CompilationError('can parse draft-04 only', schema)

    loader = loader or {}
    loader[uri] = deepcopy(schema)

    return compile(schema, uri, loader)

class Factory(object):

    def __init__(self, loader=None, spec=None):
        self.loader = loader or {}
        self.spec = spec or 'http://json-schema.org/draft-04/schema#'
        self.compilers = {
            'http://json-schema.org/draft-04/schema#': compile,
        }

    def __call__(self, schema, uri, spec=None):

        spec = schema.get('$schema', spec or self.spec)

        try:
            compiler = self.compilers[spec]
        except KeyError:
            raise CompilationError('spec {!r} not implemented'.format(spec), schema)

        self.loader[uri] = deepcopy(schema)

        return compile(schema, uri, self)

    def get(self, document):
        return self.loader.get('{}#'.format(document))

    def load(self, uri, spec=None):
        doc, _, pointer = uri.partition('#')
        doc = doc or '<document>'
        fragment = self.loader.get('{}#'.format(doc))
        pointer = pointer.lstrip('/')
        while pointer:
            member, _, pointer = pointer.partition('/')
            fragment = fragment[member]
        return self.__call__(fragment, uri, self.spec)



"""
    json.reference
    ~~~~~~~~~~~~~~

    A JSON Reference is a JSON object, which contains a member named
    "$ref", which has a JSON string value.  Example:

    { "$ref": "http://example.com/example.json#/foo/bar" }

    If a JSON value does not have these characteristics, then it SHOULD
    NOT be interpreted as a JSON Reference.

"""

from __future__ import absolute_import, print_function, unicode_literals

__all__ = ['resolve', 'Registry', 'LocalRegistry', 'NotFound', 'Forbidden']

from .bases import Registry, LocalRegistry
from .exceptions import NotFound, Forbidden
from json.pointer import DocumentPointer


def resolve(obj, pointer, registry=None):
    """resolve a local object

    :param obj: the local object.
    :param pointer: the pointer
    :param registry: the registry.
                    It mays be omited if inner json references
                    document don't refer to other documents.
    """

    registry = LocalRegistry(obj, registry or {})
    local = DocumentPointer(pointer)

    if local.document:
        registry[local.document] = obj
    local.document = '<local>'
    return registry.resolve(local)
